# coding: utf-8
import collections
import importlib
import logging
import re
import warnings

from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.utils import inspect

from modernrpc import modernrpc_settings
from modernrpc.handlers import XMLRPC, JSONRPC

logger = logging.getLogger(__name__)
warnings.simplefilter('once', DeprecationWarning)

RPC_REGISTRY_KEY = '__rpc_registry__'
DEFAULT_REGISTRY_TIMEOUT = None
ALL = "__all__"

REQUEST_KEY = 'request'
ENTRY_POINT_KEY = 'entry_point'
PROTOCOL_KEY = 'protocol'
HANDLER_KEY = 'handler'

PARAM_REXP = r':param ([\w]+):\s?(.*)'
RETURN_REXP = r':return:\s?(.*)'
PARAM_TYPE_REXP = r':type ([\w]+):\s?(.*)'
RETURN_TYPE_REXP = r':rtype:\s?(.*)'


class RPCMethod(object):

    def __init__(self, function, external_name, entry_point=ALL, protocol=ALL):
        self.module = function.__module__
        self.func_name = function.__name__
        self.external_name = external_name
        self.entry_point = entry_point
        self.protocol = protocol

        # Contains the signature of the method, as returned by "system.methodSignature"
        self.signature = []
        # Contains the method's docstring, in HTML form
        self.html_doc = ''
        # Contains doc about argumetns and their type. We store this in an ordered dict, so the args documentation
        # keep the order defined in docstring
        self.args_doc = collections.OrderedDict()
        # Contains doc about return type and return value
        self.return_doc = {}

        self.args = inspect.get_func_args(function)
        raw_docstring = self.parse_docstring(function.__doc__)
        self.html_doc = self.to_html(raw_docstring)

        # Flag the method to accept additional kwargs dict
        self.accept_kwargs = inspect.func_accepts_kwargs(function)

    @property
    def name(self):
        return self.external_name

    def __repr__(self):
        return 'RPC Method ' + self.name

    def parse_docstring(self, content):
        """
        Parse the given full docstring, and extract method description, arguments, and return documentation.

        This method try to find arguments description and types, and put the information in "args_doc" and "signature"
        members. Also parse return type and description, and put the information in "return_doc" member.
        All other lines are added to the returned string
        :param content: The full docstring
        :type content: str
        :return: The parsed method description
        :rtype: str
        """
        raw_docstring = ''

        if content is None:
            return raw_docstring

        lines = content.split('\n')
        for line in lines:
            sline = line.strip()

            param_match = re.match(PARAM_REXP, sline)
            return_match = re.match(RETURN_REXP, sline)
            param_type_match = re.match(PARAM_TYPE_REXP, sline)
            return_type_match = re.match(RETURN_TYPE_REXP, sline)

            if param_match:
                param_name, description = param_match.group(1, 2)
                if param_name == 'kwargs':
                    continue
                doc = self.args_doc.get(param_name, {})
                doc['text'] = description
                self.args_doc[param_name] = doc

            elif param_type_match:
                param_name, param_type = param_type_match.group(1, 2)
                if param_name == 'kwargs':
                    continue
                doc = self.args_doc.get(param_name, {})
                doc['type'] = param_type
                self.args_doc[param_name] = doc
                self.signature.append(param_type)

            elif return_match:
                return_description = return_match.group(1)
                self.return_doc['text'] = return_description

            elif return_type_match:
                return_description = return_type_match.group(1)
                self.return_doc['type'] = return_description
                self.signature.insert(0, return_description)

            else:
                # Add the line to help text
                raw_docstring += line + '\n'

        return raw_docstring

    def to_html(self, docstring):

        if not docstring:
            return ''

        if modernrpc_settings.MODERNRPC_DOC_FORMAT.lower() in ('rst', 'reStructred', 'reStructuredText'):
            from docutils.core import publish_parts
            return publish_parts(docstring, writer_name='html')
        elif modernrpc_settings.MODERNRPC_DOC_FORMAT.lower() in ('md', 'markdown'):
            import markdown
            return markdown.markdown(docstring)
        else:
            return "<p>{}</p>".format(docstring.replace('\n\n', '</p><p>'))

    def execute(self, *args, **kwargs):
        """
        Call the function encapsulated by the current instance

        :return:
        """
        # Try to load the method address
        module = importlib.import_module(self.module)
        func = getattr(module, self.func_name)

        # Call the rpc method, as standard python function
        if self.accept_kwargs:
            return func(*args, **kwargs)
        else:
            return func(*args)

    def __eq__(self, other):
        return \
            self.external_name == other.external_name and \
            self.module == other.module and \
            self.func_name == other.func_name and \
            self.entry_point == other.entry_point and \
            self.protocol == other.protocol

    def available_for_protocol(self, protocol):
        if self.protocol == ALL or protocol == ALL:
            return True
        else:
            valid_protocols = self.protocol if isinstance(self.protocol, list) else [self.protocol]
            return protocol in valid_protocols

    def available_for_entry_point(self, entry_point):
        if self.entry_point == ALL or entry_point == ALL:
            return True
        else:
            valid_entry_points = self.entry_point if isinstance(self.entry_point, list) else [self.entry_point]
            return entry_point in valid_entry_points

    def is_valid_for(self, entry_point, protocol):
        return self.available_for_entry_point(entry_point) and self.available_for_protocol(protocol)

    # Helpers to simplify templates generation
    def is_doc_available(self):
        """Returns True if a textual documentation is available for this method"""
        return bool(self.html_doc)

    def is_return_doc_available(self):
        """Returns True if this method's return is documented"""
        return bool(self.return_doc and (self.return_doc.get('text') or self.return_doc.get('type')))

    def is_args_doc_available(self):
        """Returns True if any of the method's arguments is documented"""
        return self.args_doc

    def is_any_doc_available(self):
        """Returns True if there is a textual documentation or a documentation on arguments or return of the method."""
        return self.is_args_doc_available() or self.is_return_doc_available() or self.is_doc_available()

    def is_available_in_json_rpc(self):
        return self.available_for_protocol(JSONRPC)

    def is_available_in_xml_rpc(self):
        return self.available_for_protocol(XMLRPC)


def get_all_method_names(entry_point=ALL, protocol=ALL, sort_methods=False):
    """"""
    # Get the current RPC registry from internal cache
    registry = cache.get(RPC_REGISTRY_KEY, default={})

    method_namess = [
        name for name, method in registry.items() if method.is_valid_for(entry_point, protocol)
    ]

    if sort_methods:
        method_namess = sorted(method_namess)

    return method_namess


def get_all_methods(entry_point=ALL, protocol=ALL, sort_methods=False):
    """Return a list of all methods in the registry supported by the given entry_point / protocol pair"""
    # Get the current RPC registry from internal cache
    registry = cache.get(RPC_REGISTRY_KEY, default={})

    if sort_methods:
        methods = [method for (_, method) in sorted(registry.items())]
    else:
        methods = registry.values()

    return [
        method for method in methods if method.is_valid_for(entry_point, protocol)
    ]


def get_method(name, entry_point, protocol):
    """Retrieve a method from the given name"""
    # Get the current RPC registry from internal cache
    registry = cache.get(RPC_REGISTRY_KEY, default={})

    # Try to find the given method in cache
    if name in registry:
        method = registry.get(name)
        # Ensure the method can be returned for given entry_point and protocol
        if method and method.is_valid_for(entry_point, protocol):
            return method

    return None


def unregister_rpc_method(method_name):
    """Remove a method from registry"""
    # Get the current RPC registry from internal cache
    registry = cache.get(RPC_REGISTRY_KEY, default={})

    if method_name in registry:
        logger.debug('Unregister RPC method {}'.format(method_name))
        del registry[method_name]

    # Update the registry in internal cache
    cache.set(RPC_REGISTRY_KEY, registry, timeout=DEFAULT_REGISTRY_TIMEOUT)


def register_rpc_method(function):
    """
    Register a function to be available as RPC method.

    The given function will be inspected to find external_name, protocol and entry_point values set by the decorator
    @rpc_method.
    :param function: A function previously decorated using @rpc_method
    :return: The name of registered method
    """
    if not getattr(function, 'modernrpc_enabled', False):
        raise ImproperlyConfigured('Error: trying to register {} as RPC method, but it has not been decorated.'
                                   .format(function.__name__))

    # Define the external name of the function
    name = getattr(function, 'modernrpc_name', function.__name__)

    logger.debug('Register RPC method {}'.format(name))

    if name.startswith('rpc.'):
        raise ImproperlyConfigured('According to RPC standard, method names starting with "rpc." are reserved for '
                                   'system extensions and must not be used. See '
                                   'http://www.jsonrpc.org/specification#extensions for more information.')

    entry_point = getattr(function, 'modernrpc_entry_point')
    protocol = getattr(function, 'modernrpc_protocol')

    # Encapsulate the function in a RPCMethod object
    method = RPCMethod(function, name, entry_point, protocol)

    # Get the current RPC registry from internal cache
    registry = cache.get(RPC_REGISTRY_KEY, default={})

    # Ensure method names are unique in the registry
    if method.external_name in registry:
        # Trying to register many times the same function is OK, because if a method is decorated
        # with @rpc_method(), it could be imported in different places of the code
        if method == registry[method.external_name]:
            return method.external_name
        # But if we try to use the same name to register 2 different methods, we
        # must inform the developer there is an error in the code
        else:
            raise ImproperlyConfigured("A RPC method with name {} has already been registered"
                                       .format(method.external_name))

    # Store the method
    registry[method.external_name] = method

    # Update the registry in internal cache
    cache.set(RPC_REGISTRY_KEY, registry, timeout=DEFAULT_REGISTRY_TIMEOUT)

    return method.external_name


def rpc_method(func=None, name=None, entry_point=ALL, protocol=ALL):
    """
    Mark a standard python function as RPC method.

    All arguments are optional

    :param func: A standard function
    :param name: Used as RPC method name instead of original function name
    :param entry_point: Default: ALL. Used to limit usage of the RPC method for a specific set of entry points
    :param protocol: Default: ALL. Used to limit usage of the RPC method for a specific protocol (JSONRPC or XMLRPC)
    :type name: str
    :type entry_point: str
    :type protocol: str
    """

    def decorated(function):

        function.modernrpc_enabled = True
        function.modernrpc_name = name or function.__name__
        function.modernrpc_entry_point = entry_point
        function.modernrpc_protocol = protocol

        return function

    # If @rpc_method is used without any argument nor parenthesis
    if func is None:
        def decorator(func):
            return decorated(func)
        return decorator
    # If @rpc_method() is used with parenthesis (with or without arguments)
    return decorated(func)


# Deprecated method: will be removed in version 1.0
def register_method(function, name=None, entry_point=ALL, protocol=ALL):
    """
    Register a function to be available as RPC method.

    All arguments but ``function`` are optional

    :param function: The python function to register
    :param name: Used as RPC method name instead of original function name
    :param entry_point: Default: ALL. Used to limit usage of the RPC method for a specific set of entry points
    :param protocol: Default: ALL. Used to limit usage of the RPC method for a specific protocol (JSONRPC or XMLRPC)
    :type function: func
    :type name: str
    :type entry_point: str
    :type protocol: str
    :return: None
    """

    warnings.warn('"register_method" is deprecated and will be removed in a future version.\nUse settings.'
                  'MODERNRPC_METHODS_MODULES to declare modules containing RPC method, and decorate each method '
                  'with @rpc_method. Refer\nto the documentation for more info.', DeprecationWarning, stacklevel=2)

    function.modernrpc_enabled = True
    function.modernrpc_name = name or function.__name__
    function.modernrpc_entry_point = entry_point
    function.modernrpc_protocol = protocol

    register_rpc_method(function)



# coding: utf-8
import inspect
import logging
import warnings
from importlib import import_module

from django.apps import AppConfig
from modernrpc.core import register_rpc_method, get_all_method_names, unregister_rpc_method
from modernrpc.modernrpc_settings import MODERNRPC_METHODS_MODULES

logger = logging.getLogger(__name__)


class ModernRpcConfig(AppConfig):

    name = 'modernrpc'
    verbose_name = 'Django Modern RPC'

    def ready(self):

        if MODERNRPC_METHODS_MODULES:

            # When project use a persistant cache (Redis, memcached, etc.) the cache is not always flushed
            # when django project is restarted. As a result, we may have a registry with a list of methods
            # from the last run.
            methods_before_lookup = get_all_method_names()

            modules_to_scan = MODERNRPC_METHODS_MODULES
            # Add internal system methods to the registry
            modules_to_scan.append('modernrpc.system_methods')

            for module_name in modules_to_scan:

                try:
                    module = import_module(module_name)

                except ImportError:
                    logger.warning('Unable to load module "{}". Please check MODERNRPC_METHODS_MODULES for invalid'
                                   ' names'.format(module_name))
                    continue

                for _, func in inspect.getmembers(module, inspect.isfunction):
                    if getattr(func, 'modernrpc_enabled', False):
                        registered_name = register_rpc_method(func)
                        try:
                            # Ensure method is not listed in methods_before_lookup after registration
                            methods_before_lookup.remove(registered_name)
                        except ValueError:
                            pass

            # Some rpc methods stored in the registry from the last run were not registered this session.
            # That means a RPC method has been deleted from the code (or the decorator has been removed)
            # In such case, we must remove the function from the registry
            for useless_method in methods_before_lookup:
                unregister_rpc_method(useless_method)

        else:
            warnings.warn("MODERNRPC_METHODS_MODULES is not set. Django-modern-rpc is not able to lookup for RPC "
                          "methos in your code. Please define settings.MODERNRPC_METHODS_MODULES to indicate "
                          "the modules that contains your methods.", category=Warning)



# coding: utf-8
import json
import logging

from django.http.response import HttpResponse
from django.utils.module_loading import import_string

from modernrpc import modernrpc_settings
from modernrpc.exceptions import RPCInternalError, RPCInvalidRequest, RPCParseError
from modernrpc.handlers.base import RPCHandler

try:
    # Python 3
    from json.decoder import JSONDecodeError
except ImportError:
    # Python 2: json.loads will raise a ValueError when loading json
    JSONDecodeError = ValueError

logger = logging.getLogger(__name__)
JSONRPC = '__json_rpc'


class JSONRPCHandler(RPCHandler):

    protocol = JSONRPC

    def __init__(self, request, entry_point):
        super(JSONRPCHandler, self).__init__(request, entry_point)
        self.request_id = None

    @staticmethod
    def valid_content_types():
        return [
            'application/json',
        ]

    def loads(self, data):
        try:
            decoder = import_string(modernrpc_settings.MODERNRPC_JSON_DECODER)
            return json.loads(data, cls=decoder)
        except JSONDecodeError as e:
            raise RPCParseError(str(e))

    def dumps(self, obj):
        try:
            encoder = import_string(modernrpc_settings.MODERNRPC_JSON_ENCODER)
            return json.dumps(obj, cls=encoder)
        except Exception:
            raise RPCInternalError('Unable to serialize result as valid JSON')

    def parse_request(self):

        encoding = self.request.encoding or 'utf-8'
        data = self.request.body.decode(encoding)
        body = self.loads(data)

        if not isinstance(body, dict):
            raise RPCParseError('Invalid request: the object must be a struct')

        if 'id' in body:
            self.request_id = body['id']
        else:
            raise RPCInvalidRequest('Missing parameter "id"')

        if 'jsonrpc' not in body:
            raise RPCInvalidRequest('Missing parameter "jsonrpc"')
        elif 'method' not in body:
            raise RPCInvalidRequest('Missing parameter "method"')
        elif 'params' not in body:
            raise RPCInvalidRequest('Missing parameter "params"')

        if body['jsonrpc'] != '2.0':
            raise RPCInvalidRequest('Invalid request. The attribute "jsonrpc" must contains "2.0"')

        return body['method'], body['params']

    @staticmethod
    def json_http_response(data):
        response = HttpResponse(data)
        response['Content-Type'] = 'application/json'
        return response

    def result_success(self, data):
        result = {
            'id': self.request_id,
            'jsonrpc': '2.0',
            'result': data,
        }
        return self.json_http_response(self.dumps(result))

    def result_error(self, exception):
        result = {
            'id': self.request_id,
            'jsonrpc': '2.0',
            'error': {
                'code': exception.code,
                'message': exception.message,
            }
        }

        return self.json_http_response(self.dumps(result))



# coding: utf-8
from django.conf import settings

#: Override this to use a different JSON serializer when decoding JSON-RPC request
MODERNRPC_JSON_DECODER = getattr(settings, 'MODERNRPC_JSON_DECODER', 'json.decoder.JSONDecoder')

#: Override this to use a different JSON deserializer when encoding JSON-RPC response
MODERNRPC_JSON_ENCODER = getattr(settings, 'MODERNRPC_JSON_ENCODER', 'django.core.serializers.json.DjangoJSONEncoder')

#: Override these to change the key used in kwars for entry point name
MODERNRPC_HANDLERS = getattr(settings, 'MODERNRPC_HANDLERS', [
    'modernrpc.handlers.JSONRPCHandler',
    'modernrpc.handlers.XMLRPCHandler',
])

#: The name set to any entry point which doesn't declare it explicitely
MODERNRPC_DEFAULT_ENTRYPOINT_NAME = getattr(settings, 'MODERNRPC_DEFAULT_ENTRYPOINT_NAME', '__default_entry_point__')

#: Set to False if you want to manipulate dates with DateTime class from Python XML RPC builtin lib.
#: If set to True, dates will be passed as datetime to the RPC function.
MODERNRPC_XML_USE_BUILTIN_TYPES = getattr(settings, 'MODERNRPC_XML_USE_BUILTIN_TYPES', True)

#: Configure the format of the docstring used to document your RPC methods. Possible values are: '', 'rst' or 'md'
MODERNRPC_DOC_FORMAT = getattr(settings, 'MODERNRPC_DOC_FORMAT', '')

#: Set the list of python modules that must be looked up to find RPC methods
MODERNRPC_METHODS_MODULES = getattr(settings, 'MODERNRPC_METHODS_MODULES', [])

#: Set to False if you want to raise an exception when a None value is passed to XML encoder
MODERNRPC_XMLRPC_ALLOW_NONE = getattr(settings, 'MODERNRPC_XMLRPC_ALLOW_NONE', True)
#: Configure the default encoding used by XML encoder
MODERNRPC_XMLRPC_DEFAULT_ENCODING = getattr(settings, 'MODERNRPC_XMLRPC_DEFAULT_ENCODING', None)



# coding: utf-8
from modernrpc.core import ENTRY_POINT_KEY, PROTOCOL_KEY, get_method, rpc_method, get_all_methods
from modernrpc.exceptions import RPCInvalidParams, RPCUnknownMethod, RPCException, RPC_INTERNAL_ERROR
from modernrpc.handlers import XMLRPC


# TODO: add missing docstring for system methods

@rpc_method(name='system.listMethods')
def __system_listMethods(**kwargs):

    entry_point = kwargs.get(ENTRY_POINT_KEY)
    protocol = kwargs.get(PROTOCOL_KEY)

    names = [method.name for method in get_all_methods(entry_point, protocol, sort_methods=True)]

    return names


@rpc_method(name='system.methodSignature')
def __system_methodSignature(method_name, **kwargs):

    entry_point = kwargs.get(ENTRY_POINT_KEY)
    protocol = kwargs.get(PROTOCOL_KEY)

    method = get_method(method_name, entry_point, protocol)
    if method is None:
        raise RPCInvalidParams('The method {} is not found in the system. Unable to retrieve signature.')
    return method.signature


@rpc_method(name='system.methodHelp')
def __system_methodHelp(method_name, **kwargs):

    entry_point = kwargs.get(ENTRY_POINT_KEY)
    protocol = kwargs.get(PROTOCOL_KEY)

    method = get_method(method_name, entry_point, protocol)
    if method is None:
        raise RPCInvalidParams('The method {} is not found in the system. Unable to retrieve method help.')
    return method.html_doc


@rpc_method(name='system.multicall', protocol=XMLRPC)
def __system_multiCall(calls, **kwargs):
    """
    Call multiple RPC methods at once.

    :param calls: An array of struct like {"methodName": string, "params": array }
    :param kwargs:
    :return:
    """
    if not isinstance(calls, list):
        raise RPCInvalidParams('method_names must be a list')

    entry_point = kwargs.get(ENTRY_POINT_KEY)
    protocol = kwargs.get(PROTOCOL_KEY)

    results = []
    for call in calls:
        method_name, params = call['methodName'], call['params']
        method = get_method(method_name, entry_point, protocol)

        try:
            if not method:
                raise RPCUnknownMethod(method_name)

            result = method.execute(*params, **kwargs)
            # From https://mirrors.talideon.com/articles/multicall.html:
            # "Notice that regular return values are always nested inside a one-element array. This allows you to
            # return structs from functions without confusing them with faults."
            results.append([result])
        except RPCException as e:
            results.append({
                'faultCode': e.code,
                'faultString': e.message,
            })
        except Exception as e:
            results.append({
                'faultCode': RPC_INTERNAL_ERROR,
                'faultString': str(e),
            })

    return results



# coding: utf-8
import logging

from django.core.exceptions import ImproperlyConfigured
from django.http.response import HttpResponse
from django.utils.decorators import method_decorator
from django.utils.module_loading import import_string
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import TemplateView

from modernrpc.core import ALL, get_method, get_all_methods
from modernrpc.exceptions import RPCInternalError, RPCException, RPCUnknownMethod, RPCInvalidParams
from modernrpc.modernrpc_settings import MODERNRPC_HANDLERS, MODERNRPC_DEFAULT_ENTRYPOINT_NAME

logger = logging.getLogger(__name__)


class RPCEntryPoint(TemplateView):
    """
    This is the main entry point class. It inherits standard Django View class.
    """

    template_name = 'modernrpc/doc_index.html'

    entry_point = MODERNRPC_DEFAULT_ENTRYPOINT_NAME
    protocol = ALL
    enable_doc = False
    enable_rpc = True

    def __init__(self, **kwargs):
        super(RPCEntryPoint, self).__init__(**kwargs)

        if not self.get_handler_classes():
            raise ImproperlyConfigured("At least 1 handler must be instantiated.")
        logger.debug('Create "{}" RPCEntryPoint view'.format(self.entry_point))

    # This disable CRSF validation for POST requests
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        """
        Overrides the default dispatch method, for 2 reasons:
         1. Filter input requests methods to respect ``enable_doc`` and ``enable_rpc`` parameters
         2. Disable CSRF validation on post request, since this is irrelevant for RPC calls.
        """

        if request.method.lower() == 'get' and not self.enable_doc:
            return self.http_method_not_allowed(request, *args, **kwargs)
        elif request.method.lower() == 'post' and not self.enable_rpc:
            return self.http_method_not_allowed(request, *args, **kwargs)

        return super(RPCEntryPoint, self).dispatch(request, *args, **kwargs)

    def _allowed_methods(self):
        allowed_methods = super(RPCEntryPoint, self)._allowed_methods()
        if not self.enable_doc:
            allowed_methods.remove('GET')
        if not self.enable_rpc:
            allowed_methods.remove('POST')
        return allowed_methods

    def get_handler_classes(self):

        handler_classes = [import_string(handler_cls) for handler_cls in MODERNRPC_HANDLERS]
        if self.protocol == ALL:
            return handler_classes
        else:
            valid_protocols = self.protocol if isinstance(self.protocol, list) else [self.protocol]
            return [cls for cls in handler_classes if cls.protocol in valid_protocols]

    def post(self, request, *args, **kwargs):
        """
        Handle any XML-RPC or JSON-RPC request.

        :param request:
        :param args:
        :param kwargs:
        :return:
        """

        logger.debug('RPC request received on entry point "{}"'.format(self.entry_point))

        for handler_cls in self.get_handler_classes():

            handler = handler_cls(request, self.entry_point)

            try:
                if not handler.can_handle():
                    continue

                logger.debug('Request will be handled by {}'.format(handler_cls.__name__))

                method, params = handler.parse_request()
                rpc_method = get_method(method, self.entry_point, handler_cls.protocol)

                if not rpc_method:
                    logger.warning('Unknown RPC method: {}'.format(method))
                    raise RPCUnknownMethod(method)

                logger.debug('RPC method {} will be executed'.format(method))

                try:
                    kwargs = {
                        'request': request,
                        'entry_point': self.entry_point,
                        'protocol': self.protocol,
                        'handler': handler,
                    }

                    result = rpc_method.execute(*params, **kwargs)

                except TypeError as e:
                    # If given arguments cannot be passed correctly to python function,
                    # raise an Invalid Params exceptions
                    raise RPCInvalidParams(str(e))

                return handler.result_success(result)

            except RPCException as e:
                return handler.result_error(e)
            except Exception as e:
                return handler.result_error(RPCInternalError(str(e)))

        logger.error('Received a request impossible to handle with available handlers.')

        return HttpResponse('Unable to handle your request. Please ensure you called the right entry point. If not, '
                            'this could be a server error.')

    def get_context_data(self, **kwargs):
        """Update context data with list of RPC methods of the current entry point"""
        ctx = super(RPCEntryPoint, self).get_context_data(**kwargs)
        ctx.update({
            'methods': get_all_methods(self.entry_point, sort_methods=True),
        })
        return ctx



# coding: utf-8

from modernrpc.handlers.jsonhandler import JSONRPC, JSONRPCHandler
from modernrpc.handlers.xmlhandler import XMLRPC, XMLRPCHandler

__all__ = [
    'JSONRPC',
    'XMLRPC',
    'JSONRPCHandler',
    'XMLRPCHandler',
]



# coding: utf-8
import logging

from django.http.response import HttpResponse

from modernrpc.exceptions import RPCParseError, RPCInvalidRequest
from modernrpc.handlers.base import RPCHandler
from modernrpc.modernrpc_settings import MODERNRPC_XML_USE_BUILTIN_TYPES, MODERNRPC_XMLRPC_ALLOW_NONE, \
    MODERNRPC_XMLRPC_DEFAULT_ENCODING

try:
    # Python 3
    import xmlrpc.client as xmlrpc_module
    from xmlrpc.client import ResponseError
except ImportError:
    # Python 2
    import xmlrpclib as xmlrpc_module
    from xmlrpclib import ResponseError

logger = logging.getLogger(__name__)
XMLRPC = '__xml_rpc'


class XMLRPCHandler(RPCHandler):

    protocol = XMLRPC

    def __init__(self, request, entry_point):
        super(XMLRPCHandler, self).__init__(request, entry_point)
        self.marshaller = xmlrpc_module.Marshaller(encoding=MODERNRPC_XMLRPC_DEFAULT_ENCODING,
                                                   allow_none=MODERNRPC_XMLRPC_ALLOW_NONE)

    @staticmethod
    def valid_content_types():
        return [
            'text/xml',
        ]

    def loads(self, data):
        try:
            try:
                return xmlrpc_module.loads(data, use_builtin_types=MODERNRPC_XML_USE_BUILTIN_TYPES)
            except TypeError:
                return xmlrpc_module.loads(data, use_datetime=MODERNRPC_XML_USE_BUILTIN_TYPES)
        except ResponseError as e:
            raise RPCInvalidRequest(e)
        except Exception as e:
            # Depending on the concrete parser used in loads, we can't determine wich exception can be raised during
            # XML loading. So we catch the generic Exception and a RpcParseError, since the error is mostly related
            # to request XML parsing.
            raise RPCParseError(e)

    def dumps(self, obj):

        # Marshaller has a specific handling of Fault instance. It is given without modification
        if isinstance(obj, xmlrpc_module.Fault):
            return self.marshaller.dumps(obj)

        # xmlrpc.client.Marshaller (or the Python equivalent) expects a list of objects to dumps.
        # It will loop over these objects, will output a <param><value>X</value></param> for each element.
        # This is not the return defined in XML-RPC standard, see http://xmlrpc.scripting.com/spec.html:
        # > The body of the response is a single XML structure, a <methodResponse>, which can contain
        # > a single <params> which contains a single <param> which contains a single <value>.
        # For any type to dump, we need to give a list of only 1 element:
        return self.marshaller.dumps([obj])

    def parse_request(self):

        encoding = self.request.encoding or 'utf-8'
        data = self.request.body.decode(encoding)

        params, method_name = self.loads(data)

        if method_name is None:
            raise RPCInvalidRequest('Missing methodName')

        return method_name, params

    @staticmethod
    def xml_http_response(data):
        response = HttpResponse(data)
        response['Content-Type'] = 'text/xml'
        return response

    def result_success(self, data):

        raw_response = '<?xml version="1.0"?>'
        raw_response += '<methodResponse>'
        raw_response += self.dumps(data)
        raw_response += '</methodResponse>'

        return self.xml_http_response(raw_response)

    def result_error(self, exception):

        raw_response = '<?xml version="1.0"?>'
        raw_response += '<methodResponse>'
        raw_response += self.dumps(xmlrpc_module.Fault(exception.code, exception.message))
        raw_response += '</methodResponse>'

        return self.xml_http_response(raw_response)



#!/usr/bin/env python
import re
from codecs import open
from os import path

# Always prefer setuptools over distutils
from setuptools import setup, find_packages


def read(*names, **kwargs):
    with open(
        path.join(path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='django-modern-rpc',
    version=find_version("modernrpc", "__init__.py"),
    description='Simple and modern JSON-RPC and XML-RPC server implementation for Django',
    long_description=read('README.rst'),
    author='Antoine Lorence',
    author_email='antoine.lorence@gmail.com',
    url='https://github.com/alorence/django-modern-rpc',
    packages=find_packages(exclude=['modernrpc.tests', 'modernrpc.tests.*']),
    # Files included in sdist (using MANIFEST.in) will also be included in bdist*
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Framework :: Django :: 1.10",
        "Framework :: Django :: 1.9",
        "Framework :: Django :: 1.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Intended Audience :: Developers",
    ],
    install_requires=[
        "Django>=1.8.0",
    ]
)
